from typing import Optional, Callable
from time import sleep
from pathlib import Path
from functools import partial
import logging
import sys

from mpris_server.server import Server

from .base import get_chromecast, Seconds, get_chromecast_via_host, \
  NoChromecastFoundException, LOG_LEVEL, get_chromecast_via_uuid, \
  DEFAULT_RETRY_WAIT, NoChromecastFoundException, RC_NO_CHROMECAST, \
  DATA_DIR, NAME, RC_NOT_RUNNING, PID, NO_DEVICE, DEFAULT_WAIT
from .adapter import ChromecastAdapter
from .listeners import register_mpris_adapter


FuncMaybe = Optional[Callable]


def set_log_level(level: str = LOG_LEVEL):
  level = level.upper()
  logging.basicConfig(level=level)


def run_safe(
  name: Optional[str],
  host: Optional[str],
  uuid: Optional[str],
  wait: Optional[float],
  retry_wait: Optional[float],
  icon: bool,
  log_level: str
):
  try:
    run_server(
      name,
      host,
      uuid,
      wait,
      retry_wait,
      icon,
      log_level
    )

  except NoChromecastFoundException as e:
    logging.warning(f"Device {e} not found")
    sys.exit(RC_NO_CHROMECAST)


def create_adapters_and_server(
  name: Optional[str],
  host: Optional[str],
  uuid: Optional[str],
  retry_wait: Optional[float] = DEFAULT_RETRY_WAIT,
) -> Optional[Server]:
  if host:
    chromecast = get_chromecast_via_host(host, retry_wait)

  elif name:
    chromecast = get_chromecast(name, retry_wait)

  elif uuid:
    chromecast = get_chromecast_via_uuid(uuid, retry_wait)

  else:
    chromecast = get_chromecast(retry_wait=retry_wait)

  if not chromecast:
    return None

  chromecast_adapter = ChromecastAdapter(chromecast)
  mpris = Server(name=chromecast.name, adapter=chromecast_adapter)
  mpris.publish()

  register_mpris_adapter(chromecast, mpris, chromecast_adapter)

  return mpris


def retry_until_found(
  name: Optional[str],
  host: Optional[str],
  uuid: Optional[str],
  wait: Optional[Seconds] = DEFAULT_WAIT,
  retry_wait: Optional[float] = DEFAULT_RETRY_WAIT,
) -> Optional[Server]:
  """
    If the Chromecast isn't found, keep trying to find it.

    If `wait` is None, then retrying is disabled.
  """

  while True:
    mpris = create_adapters_and_server(name, host, uuid, retry_wait)

    if mpris or wait is None:
      return mpris

    logging.info(f"{name or NO_DEVICE} not found. Waiting {wait} seconds before retrying.")
    sleep(wait)


def run_server(
  name: Optional[str],
  host: Optional[str],
  uuid: Optional[str],
  wait: Optional[float] = DEFAULT_WAIT,
  retry_wait: Optional[float] = DEFAULT_RETRY_WAIT,
  icon: bool = False,
  log_level: str = LOG_LEVEL
):
  set_log_level(log_level)
  mpris = retry_until_found(name, host, uuid, wait, retry_wait)

  if mpris and icon:
    mpris.adapter.wrapper.set_icon(True)

  if not mpris:
    raise NoChromecastFoundException(name)

  mpris.loop()



from typing import List

from pychromecast import Chromecast

from mpris_server.adapters import Metadata, PlayState, MprisAdapter, \
  Microseconds, VolumeDecimal, RateDecimal
from mpris_server.base import URI, MIME_TYPES, DEFAULT_RATE, DbusObj, \
  Track

from .wrapper import ChromecastWrapper


class ChromecastAdapter(MprisAdapter):
  def __init__(self, chromecast: Chromecast):
    self.wrapper = ChromecastWrapper(chromecast)
    super().__init__(self.wrapper.name)

  def get_uri_schemes(self) -> List[str]:
    return URI

  def get_mime_types(self) -> List[str]:
    return MIME_TYPES

  def can_quit(self) -> bool:
    return self.wrapper.can_quit()

  def can_go_next(self) -> bool:
    return self.wrapper.can_play_next()

  def can_go_previous(self) -> bool:
    return self.wrapper.can_play_prev()

  def can_play(self) -> bool:
    return self.wrapper.can_play()

  def can_pause(self) -> bool:
    return self.wrapper.can_pause()

  def can_seek(self) -> bool:
    return self.wrapper.can_seek()

  def can_control(self) -> bool:
    return self.wrapper.can_control()

  def quit(self):
    self.wrapper.quit()

  def get_current_position(self) -> Microseconds:
    return self.wrapper.get_current_position()

  def next(self):
    self.wrapper.next()

  def previous(self):
    self.wrapper.previous()

  def pause(self):
    self.wrapper.pause()

  def resume(self):
    self.play()

  def stop(self):
    self.wrapper.stop()

  def play(self):
    self.wrapper.play()

  def get_playstate(self) -> PlayState:
    return self.wrapper.get_playstate()

  def seek(self, time: Microseconds):
    self.wrapper.seek(time)

  def open_uri(self, uri: str):
    self.wrapper.open_uri(uri)

  def is_repeating(self) -> bool:
    return self.wrapper.is_repeating()

  def is_playlist(self) -> bool:
    return self.wrapper.is_playlist()

  def set_repeating(self, val: bool):
    pass

  def set_loop_status(self, val: str):
    pass

  def get_rate(self) -> RateDecimal:
    return self.wrapper.get_rate()

  def set_rate(self, val: RateDecimal):
    pass

  def get_shuffle(self) -> bool:
    return False

  def set_shuffle(self, val: bool):
    return False

  def get_art_url(self, track: int = None) -> str:
    return self.wrapper.get_art_url(track)

  def get_volume(self) -> VolumeDecimal:
    return self.wrapper.get_volume()

  def set_volume(self, val: VolumeDecimal):
    self.wrapper.set_volume(val)

  def is_mute(self) -> bool:
    return self.wrapper.is_mute()

  def set_mute(self, val: bool):
    self.wrapper.set_mute(val)

  def get_stream_title(self) -> str:
    return self.wrapper.get_stream_title()

  def get_previous_track(self) -> Track:
    pass

  def get_next_track(self) -> Track:
    pass

  def get_duration(self) -> Microseconds:
    return self.wrapper.get_duration()

  def metadata(self) -> Metadata:
    return self.wrapper.metadata()

  def get_current_track(self) -> Track:
    return self.wrapper.get_current_track()

  def get_desktop_entry(self) -> str:
    return self.wrapper.get_desktop_entry()

  def add_track(
    self,
    uri: str,
    after_track: DbusObj,
    set_as_current: bool
  ):
    self.wrapper.add_track(uri, after_track, set_as_current)

  def can_edit_track(self) -> bool:
    return self.wrapper.can_edit_track()



from typing import Optional
import logging

import click

from .base import LOG_LEVEL, DEFAULT_RETRY_WAIT
from .run import run_safe


DEPRECATION_WARNING: str = \
  "The chromecast_mpris command is deprecated. Please run cast_control instead."

HELP: str = f"""
Control casting devices via Linux media controls and desktops.

This daemon connects your casting device directly to the D-Bus media player interface.

NOTE: {DEPRECATION_WARNING}
"""


@click.command(
  help=HELP
)
@click.option('--name', '-n',
  default=None, show_default=True, type=click.STRING,
  help="Connect to a device via its name, otherwise control the first device found.")
@click.option('--host', '-h',
  default=None, show_default=True, type=click.STRING,
  help="Connect to a device via its hostname or IP address, otherwise control the first device found.")
@click.option('--uuid', '-u',
  default=None, show_default=True, type=click.STRING,
  help="Connect to a device via its UUID, otherwise control the first device found.")
@click.option('--wait', '-w',
  default=None, show_default=True, type=click.FLOAT,
  help="Seconds to wait between trying to make initial successful connections to a device.")
@click.option('--retry-wait', '-r',
  default=DEFAULT_RETRY_WAIT, show_default=True, type=click.FLOAT,
  help="Seconds to wait between reconnection attempts if a successful connection is interrupted.")
@click.option('--icon', '-i',
  is_flag=True, default=False, show_default=True, type=click.BOOL,
  help="Use a lighter icon instead of the dark icon. The lighter icon goes well with dark themes.")
@click.option('--log-level', '-l',
  default=LOG_LEVEL, show_default=True, type=click.STRING,
  help='Set the debugging log level.')
def cmd(
  name: Optional[str],
  host: Optional[str],
  uuid: Optional[str],
  wait: Optional[float],
  retry_wait: Optional[float],
  icon: bool,
  log_level: str
):
  logging.warn(DEPRECATION_WARNING)

  run_safe(
    name,
    host,
    uuid,
    wait,
    retry_wait,
    icon,
    log_level
  )


if __name__ == "__main__":
  cmd()
